from patas.schemas import load_cluster, load_experiment, format_as_yaml

def unindent(n, data):

    prefix = ' ' * n
    lines = [x[n:] if x.startswith(prefix) else x for x in data.split('\n')]
    return '\n'.join(lines[1:] if lines[0] == '' else lines)


def test_cluster_schema():
    cluster = load_cluster('./tests/data/cluster.yaml')
    assert cluster.name == 'wespa'
    assert len(cluster.nodes) == 2

def test_experiment_schema():
    experiment = load_experiment('./tests/data/experiment.yaml')
    assert experiment.name == 'trackerx'
    assert experiment.workdir == '$HOME/Sources/patas/test'
    assert experiment.repeat == 3
    assert experiment.max_tries == 2
    assert experiment.redo == True
    assert len(experiment.cmd) == 1
    assert len(experiment.vars) == 3

def test_format_as_yaml():
    expected_output = unindent(8, """
        cmd:
        - ls -la {v1}
        max_tries: 2
        name: trackerx
        redo: true
        repeat: 3
        vars:
        - type: list
          values:
          - a
          - b
          - c
          - d
        - factor: 1
          max: 10
          min: 1
          type: arithmetic
        - factor: 2
          max: 17
          min: 1
          type: geometric
        workdir: $HOME/Sources/patas/test
        """)

    experiment = load_experiment('./tests/data/experiment.yaml')
    assert format_as_yaml(experiment) == expected_output




from patas.utils import expand_path, abort, warn
from glob import glob

import yaml
import csv
import os
import re


class Pattern:
    
    def __init__(self, name, pattern):

        self.name  = "out_" + name
        self.raw   = pattern
        self.data  = None

        self.raw   = self.raw.replace("@float@", "[-0-9\\.e\\+]+")
        self.raw   = self.raw.replace("@int@", "[-0-9\\+]+")
        self.raw   = self.raw.replace("@slug@", "[-0-9a-zA-Z_]+")
        self.raw   = self.raw.replace("@str@", "\"[^\"]+\"")

        self.regex = re.compile(self.raw)
   
    def clear(self):

        self.data = None

    def get_name(self):

        return self.name

    def get_value(self):

        # if self.data is None:
        #     sys.stdout.write("!(" + self.name + ")")
        return self.data

    def value(self):

        return self.data
    
    def check(self, row):

        m:re.Match = self.regex.search(row)

        if m:
            self.data = m.groups()[0]
            return True
        else:
            return False


class TaskParser:

    def __init__(self, experiment_info_filepath, patterns, linebreakers, verbose=False):

        self.linebreakers = linebreakers
        self.patterns = patterns
        self.verbose = verbose
        self.header_names = []
        self.header_map = {}

        # Add input variable columns

        with open(experiment_info_filepath, "r") as fin:
            experiment_info = yaml.load(fin, Loader=yaml.FullLoader)
        
        for var_name in experiment_info["variables"].keys():
            self._add_column("in_" + var_name)

        # Add pattern columns

        for pattern in patterns:
            self._add_column(pattern.name)

        # Add columns with extra info
        
        if verbose:
            for var_name in ["break_id", "task_id", "repeat_id", "combination_id", "experiment_id", "experiment_name", 
                            "duration", "started_at", "ended_at", "tries", "max_tries", 
                            "cluster_id", "cluster_name", "node_id", "node_name", "worker_id", "output_dir", "work_dir"]:

                self._add_column(var_name)

        self.row_values = [None for _ in range(len(self.header_map))]


    def _add_column(self, name):

        if name in self.header_map:
            abort("Header appears multiple times:", name)
        
        self.header_names.append(name)
        self.header_map[name] = len(self.header_map)


    def digest(self, task_folderpath, writer):

        # Filepaths for each task files

        info_filepath   = os.path.join(task_folderpath, "info.yml")
        output_filepath = os.path.join(task_folderpath, "stdout")
        done_filepath   = os.path.join(task_folderpath, ".done")

        # We clean the output row, as we will populate it with the data from the task

        for i in range(len(self.row_values)):
            self.row_values[i] = None

        # If the done file does not exist, it means the task wasnot completed, this is likely a user mistake

        if not os.path.exists(done_filepath):
            warn("Task not completed:", task_folderpath)
        
        # Load the task info

        with open(info_filepath, "r") as fin:
            data = yaml.load(fin, Loader=yaml.FullLoader)
        
        # Load the variable values

        for name, value in data["combination"].items():
            name = "in_" + name
            self.row_values[self.header_map[name]] = value

        # Load the extra data

        break_idd = 0

        if self.verbose:

            self.row_values[self.header_map["break_id"        ]] = str(break_idd)
            self.row_values[self.header_map["task_id"         ]] = data["task_id"         ]
            self.row_values[self.header_map["repeat_id"       ]] = data["repeat_id"       ]
            self.row_values[self.header_map["combination_id"  ]] = data["combination_id"  ]
            self.row_values[self.header_map["experiment_id"   ]] = data["experiment_id"   ]
            self.row_values[self.header_map["experiment_name" ]] = data["experiment_name" ]
            self.row_values[self.header_map["duration"        ]] = data["duration"        ]
            self.row_values[self.header_map["started_at"      ]] = data["started_at"      ]
            self.row_values[self.header_map["ended_at"        ]] = data["ended_at"        ]
            self.row_values[self.header_map["tries"           ]] = data["tries"           ]
            self.row_values[self.header_map["max_tries"       ]] = data["max_tries"       ]
            self.row_values[self.header_map["output_dir"      ]] = data["output_dir"      ]
            self.row_values[self.header_map["work_dir"        ]] = data["work_dir"        ]

            self.row_values[self.header_map["cluster_id"  ]] = data["env_variables"]["PATAS_CLUSTER_ID"  ]
            self.row_values[self.header_map["cluster_name"]] = data["env_variables"]["PATAS_CLUSTER_NAME"]
            self.row_values[self.header_map["node_id"     ]] = data["env_variables"]["PATAS_NODE_ID"     ]
            self.row_values[self.header_map["node_name"   ]] = data["env_variables"]["PATAS_NODE_NAME"   ]
            self.row_values[self.header_map["worker_id"   ]] = data["env_variables"]["PATAS_WORKER_ID"   ]

        # Parse the patterns and write to output

        pattern:Pattern = None

        with open(output_filepath, "r") as fin:
            for line in fin:

                # Check all patterns and capture the desired variables

                for pattern in self.patterns:
                    if pattern.check(line):
                        self.row_values[self.header_map[pattern.get_name()]] = pattern.get_value()
                
                # If line breaks are defined, we will write a row every time one of them is detected and clean the captured patterns so far

                for linebreaker in self.linebreakers:
                    if linebreaker.check(line):

                        writer.writerow(self.row_values)

                        break_idd += 1
                        self.row_values[self.header_map["break_id"]] = str(break_idd)

                        for pattern in self.patterns:
                            self.row_values[self.header_map[pattern.get_name()]] = None
        
        # If we are not using line breakers, we always write the row after all lines were processed

        if not self.linebreakers:
            writer.writerow(self.row_values)


class ExperimentParser:

    def __init__(self, patterns, linebreakers, verbose=False):

        self.linebreakers = linebreakers
        self.patterns     = patterns
        self.verbose      = verbose


    def start(self, experiment_folder, output_file):

        experiment_folder        = expand_path(experiment_folder)
        experiment_info_filepath = os.path.join(experiment_folder, "info.yml")
        task_folderpaths         = glob(os.path.join(experiment_folder, "*"))
        parser                   = TaskParser(experiment_info_filepath, self.patterns, self.linebreakers, self.verbose)
        
        if output_file is None:
            
            experiment_name = os.path.basename(os.path.dirname(experiment_info_filepath))
            output_file     = os.path.join(experiment_folder, experiment_name + ".csv")

        with open(output_file, "w") as fout:

            writer = csv.writer(fout, delimiter=";")
            writer.writerow(parser.header_names)

            for task_folderpath in task_folderpaths:
                if os.path.isdir(task_folderpath):
                    parser.digest(task_folderpath, writer)



#!/usr/bin/env python3

try:
    from . import consts as c
except:
    from patas import consts as c

from patas.schemas import Cluster, Experiment, Node, ListVariable, ArithmeticVariable, GeometricVariable, load_cluster, load_experiment
from patas.utils import error, node_cpu_count, abort
from patas.parse import ExperimentParser, Pattern
from patas.query_engine import QueryEngine
from patas.grid_exec import GridExec

from multiprocessing import cpu_count

import argparse
import sys

DEFAULT_PATAS_OUTPUT_DIR = './patasout'

def show_main_syntax():
    print("Syntax: patas [explore,parse,query,plot] {ARGS}")

def show_explore_syntax():
    print("Syntax: patas explore [grid,cdeepso] {ARGS}")

def do_version(args):
    print(f"Patas {c.version}")

def do_explore_grid(argv):
    
    # argparse for 'patas explore grid'

    parser = argparse.ArgumentParser(
                        prog='patas explore grid',
                        description='Execute a program permutating its input parameters.',
                        epilog="Check the README.md to learn more tips on how to use this feature: https://github.com/diegofps/patas/blob/main/README.md",
                        formatter_class=argparse.RawDescriptionHelpFormatter)

    # General parameters

    parser.add_argument('--cluster',
                        type=str,
                        metavar='FILEPATH',
                        dest='cluster',
                        help="path to a cluster file",
                        action='append')

    parser.add_argument('--experiment',
                        type=str,
                        metavar='FILEPATH',
                        dest='experiment',
                        help="path to an experiment file",
                        action='append')

    parser.add_argument('--redo',
                        dest='redo_tasks',
                        help="forces patas to redo all tasks when an experiment is executed again",
                        action='store_true')

    parser.add_argument('--recreate',
                        dest='recreate',
                        help="recreate the entire output folder if it contains a different experiment configuration",
                        action='store_true')

    parser.add_argument('--filter-tasks',
                        type=str,
                        default=[],
                        nargs='*',
                        metavar='A:B',
                        dest='task_filters',
                        help="restricts the tasks that will be executed [A:B, A:, :B, :]",
                        action='append')

    parser.add_argument('--filter-nodes',
                        type=str,
                        default=[],
                        metavar='TAG',
                        nargs='*',
                        dest='node_filters',
                        help="filter nodes that match these tags (AND)",
                        action='store')

    parser.add_argument('--node',
                        type=str,
                        nargs='*',
                        metavar='NAME USER@HOST:PORT WORKERS TAG1 ...',
                        dest='node',
                        help="adds a machine with the given number of workers to the cluster",
                        action='append')

    parser.add_argument('-y',
                        dest='confirmed',
                        help="skip confirmation before starting the tasks",
                        action='store_true')

    parser.add_argument('-o',
                        type=str,
                        default=DEFAULT_PATAS_OUTPUT_DIR,
                        metavar='FOLDER',
                        dest='output_folder',
                        help="folder to store the program outputs",
                        action='store')

    # Quick Experiment parameters

    parser.add_argument('--name',
                        type=str,
                        default='grid',
                        metavar='NAME',
                        dest='name',
                        help="changes the name of the experiment folder, default is grid",
                        action='store')

    parser.add_argument('--vl',
                        type=str,
                        nargs='*',
                        default=[],
                        metavar='VAL',
                        dest='var_list',
                        help="defines an input variable that can assume a fixed number of values",
                        action='append')

    parser.add_argument('--va',
                        type=str,
                        nargs=4,
                        default=[],
                        metavar=('NAME', 'MIN', 'MAX', 'FACTOR'),
                        dest='var_arithmetic',
                        help="defines an input variable that grows according to an arithmetic progression",
                        action='append')

    parser.add_argument('--vg',
                        type=str,
                        nargs=4,
                        default=[],
                        metavar=('NAME', 'MIN', 'MAX', 'FACTOR'),
                        dest='var_geometric',
                        help="defines an input variable that grows according to a geometric progression",
                        action='append')

    parser.add_argument('--repeat',
                        type=int,
                        metavar='R',
                        default='1',
                        dest='repeat',
                        help="number of times that each combination must be executed",
                        action='store')

    parser.add_argument('--max-tries',
                        type=int,
                        metavar='T',
                        dest='max_tries',
                        help="maximum number of retries after a task has failed",
                        action='store')

    parser.add_argument('--workdir',
                        type=str,
                        metavar='W',
                        dest='workdir',
                        help="The working directory in the remote machines that will start the tasks",
                        action='store')

    parser.add_argument('--cmd',
                        type=str,
                        metavar='CMD',
                        required=True,
                        dest='cmd',
                        help="command to be executed. Use {VAR_NAME} to replace its parameters with a named variable",
                        action='append')

    args        = parser.parse_args(args=argv)
    experiments = []
    clusters    = []

    # If experiment parameters are provided, create an experiment for them

    experiment = Experiment()

    if args.name:
        experiment.name = args.name

    if args.repeat:
        experiment.repeat = args.repeat
    
    if args.workdir:
        experiment.workdir = args.workdir
    else:
        from os import environ
        experiment.workdir = environ['PWD']
    
    if args.cmd:
        experiment.cmd = args.cmd

    if args.max_tries:
        experiment.max_tries = args.max_tries

    for values in args.var_list:
        var = ListVariable()
        var.name = values[0]
        var.values = values[1:]
        experiment.vars.append(var)

    for name, min, max, factor in args.var_arithmetic:
        var = ArithmeticVariable()
        var.name = name
        var.min = min
        var.max = max
        var.factor = factor
        var.update()
        experiment.vars.append(var)

    for name, min, max, factor in args.var_geometric:
        var = GeometricVariable()
        var.name = name
        var.min = min
        var.max = max
        var.factor = factor
        var.update()
        experiment.vars.append(var)

    if experiment.cmd or experiment.vars:
        experiments.append(experiment)

    # If experiment files are provided, load all of them

    if args.experiment:
        for filepath in args.experiment:
            experiment = load_experiment(filepath)
            experiments.append(experiment)
    
    # If cluster files are provided, we will use them

    if args.cluster:
        for filepath in args.cluster:
            print(filepath)
            cluster = load_cluster(filepath)
            clusters.append(cluster)
    
    # If nodes are provided, we will add them to the QuickCluster

    if args.node:
        
        cluster = Cluster()
        cluster.name = 'cluster'
        clusters.append(cluster)

        for i, node_params in enumerate(args.node):

            node = Node()
            node.name = f'node{i}'

            address      = node_params[0]
            node.workers = int(node_params[1]) if len(node_params) > 1 else None
            node.tags    = node_params[2:] if len(node_params) > 2 else []

            if ':' in address:
                address, port = address.split(':', 1)
                node.port = int(port)
            
            if '@' in address:
                node.user, node.hostname = address.split('@', 1)
                
            else:
                node.hostname = address

            if node.workers is None:
                node.workers = node_cpu_count(node.user, node.hostname, node.port, 1)
                
            cluster.nodes.append(node)

    # If no cluster or node is provided, we will create a simple one for the local machine

    if not clusters:

        node          = Node()
        node.name     = 'localhost'
        node.hostname = 'localhost'
        node.workers  = cpu_count()

        cluster       = Cluster()
        cluster.name  = 'cluster'

        cluster.nodes.append(node)

        clusters.append(cluster)

    # Prepare the tasks_filter

    task_filters = []

    for filters in args.task_filters:
        for task in filters:
            try:

                cells = task.split(':')

                if len(cells) == 2:
                    i1 = int(cells[0]) if cells[0] else 0
                    i2 = int(cells[1]) if cells[1] else float('inf')
                    task_filters.append((i1, i2))

                elif len(cells) == 1:
                    i1 = int(cells[0]) if cells[0] else 0
                    task_filters.append((i1, i1+1))

                else:
                    error(f'Invalid attribute for --task-filter: {task}')

            except ValueError:
                error(f'Invalid attribute for --task-filter: {task}')

    # Prepare the node_filters

    node_filters = [ x for filters in args.node_filters for x in filters ]

    # Create and run GridExec

    burn = GridExec(task_filters, node_filters, args.output_folder, args.redo_tasks, args.recreate, args.confirmed, experiments, clusters)
    burn.start()

def do_parse(argv):
    
    # argparse for 'patas parse'

    parser = argparse.ArgumentParser(
                        prog='patas parse',
                        description='Parse the output files from patas exec and generate a summary in csv',
                        epilog="Check the README.md to learn more tips on how to use this feature: https://github.com/diegofps/patas/blob/main/README.md",
                        formatter_class=argparse.RawDescriptionHelpFormatter)

    # General parameters

    parser.add_argument('-e',
                        type=str,
                        metavar='FOLDERPATH',
                        dest='experiment_folder',
                        required=True,
                        help="path to the experiment folder that must be parsed",
                        action='store')

    parser.add_argument('-o',
                        type=str,
                        metavar='FILEPATH',
                        dest='output_file',
                        help="path to the output csv file",
                        action='store')

    parser.add_argument('-p',
                        type=str,
                        metavar=('OUT_NAME', 'REGEX'),
                        default=[],
                        nargs=2,
                        dest='patterns',
                        help="regex containing a single group capture indicating the data that must be captured",
                        action='append')

    parser.add_argument('-n',
                        type=str,
                        metavar='REGEX',
                        default=[],
                        dest='linebreakers',
                        help="emit a line break in the output csv",
                        action='append')

    args = parser.parse_args(args=argv)

    # Convert the input values from '-p' and '-n' into Pattern objects

    patterns     = [Pattern(name, pattern) for name, pattern in args.patterns]
    linebreakers = [Pattern(None, pattern) for pattern in args.linebreakers]

    # Parse the folder and generate the output file

    parser = ExperimentParser(patterns, linebreakers, True)
    parser.start(args.experiment_folder, args.output_file)

def do_query(argv):
    
    # argparse for 'patas query'

    parser = argparse.ArgumentParser(
                        prog='patas query',
                        description='Execute sql queries in the parsed experiments',
                        epilog="Check the README.md to learn more tips on how to use this feature: https://github.com/diegofps/patas/blob/main/README.md",
                        formatter_class=argparse.RawDescriptionHelpFormatter)

    # General parameters

    parser.add_argument('-i',
                        type=str,
                        default=DEFAULT_PATAS_OUTPUT_DIR,
                        metavar='PATASPATH',
                        dest='patas_folder',
                        help="path to the output folder containing the experiments",
                        action='store')

    parser.add_argument('-p',
                        dest='pretty_print',
                        help="pretty print the output",
                        action='store_true')

    parser.add_argument(
                        type=str,
                        metavar='QUERY',
                        dest='query',
                        help="query that must be executed",
                        action='store')
    
    args = parser.parse_args(args=argv)

    engine = QueryEngine(args.patas_folder)
    engine.query(args.query, args.pretty_print)


def do_doctor(args):
    abort("doctor is not implemented yet.")

def do_explore_cdeepso(args):
    abort("CDEEPSO is not implemented yet.")

def do_plot(args):
    abort("Plotting is not implemented yet.")


def main(*params):

    args = sys.argv[1:]

    if len(args) == 0:
        show_main_syntax()
    
    elif args[0] == 'explore':

        if len(args) == 1:
            show_explore_syntax()
        
        elif args[1] == 'grid':
            do_explore_grid(args[2:])

        elif args[1] == 'cdeepso':
            do_explore_cdeepso(args[2:])

        else:
            show_explore_syntax()

    elif args[0] == 'parse':
        do_parse(args[1:])

    elif args[0] == 'query':
        do_query(args[1:])

    elif args[0] == 'plot':
        do_doctor(args[1:])

    elif args[0] == 'doctor':
        do_doctor(args[1:])

    elif args[0] == '-v' or args[0] == '--version':
        do_version()

    elif args[0] == '-h' or args[0] == '--help':
        show_main_syntax()

    else:
        show_main_syntax()


if __name__ == "__main__":
    main()



from .utils import error

import yaml


def format_as_dict(obj):

    if isinstance(obj, Schema):
        obj = obj.__dict__
    
    if isinstance(obj, list):
        return [format_as_dict(x) for x in obj]

    if isinstance(obj, tuple):
        return (format_as_dict(x) for x in obj)
    
    if isinstance(obj, dict):
        return { k:format_as_dict(v) for k,v in obj.items() if not k.startswith('_') }

    return obj


class Schema:

    def load_property(self, name, data, mandatory=False):
        
        if name in data:
            setattr(self, name, data[name])
        
        elif mandatory:
            error("Missing mandatory property in Node schema: " + name)


class Node(Schema):

    def __init__(self, data=None):
        
        self.tags = []
        self.hostname = ""
        self.workers = 1
        self.user = None
        self.port = 22
        self.name = 'noname'
        self.private_key = None
        self._credential = None

        if data is not None:
            self.init_from(data)

    def init_from(self, data):

        self.load_property('hostname', data, mandatory=True)
        self.load_property('workers', data)
        self.load_property('tags', data)
        self.load_property('user', data)
        self.load_property('port', data)
        self.load_property('name', data)

        return self

    @property
    def credential(self):
        
        if self._credential is None:
        
            if self.user is None:
                self._credential = self.ip
        
            else:
                self._credential = self.user + "@" + self.ip

        return self._credential
    
    @property
    def ip(self):
        
        return self.hostname


class Cluster(Schema):

    def __init__(self, data=None):
        
        self.name = "default"
        self.nodes = []

        if data is not None:
            self.init_from(data)

    def init_from(self, data):

        self.load_property('name', data)
        self.load_property('nodes', data, mandatory=True)
        
        self.nodes = [Node(x) for x in self.nodes]

        return self


class ListVariable(Schema):
    
    def __init__(self, data=None):

        self.type = 'list'
        self.name = None
        self.values = []

        if data is not None:
            self.init_from(data)
    
    def init_from(self, data):

        self.load_property('values', data, mandatory=True)
        self.load_property('name', data)
        self.load_property('type', data)

        return self


class ArithmeticVariable(Schema):
    
    def __init__(self, data=None):

        self.type = 'arithmetic'
        self.name = None
        self.factor = 1
        self.min = 0
        self.max = 10

        if data is not None:
            self.init_from(data)

        self.update()
    
    def update(self):
        self._values = self.arange(self.first, self.last, self.step)
    
    @property
    def values(self):
        return self._values
    
    def init_from(self, data):

        self.load_property('factor', data)
        self.load_property('name', data)
        self.load_property('min', data)
        self.load_property('max', data)

        return self


class GeometricVariable(Schema):
    
    def __init__(self, data=None):

        self.type = 'geometric'
        self.name = None
        self.factor = 2
        self.min = 1
        self.max = 17

        if data is not None:
            self.init_from(data)
    
        self.update()
    
    def update(self):
        self._values = list()
        current = self.first
        
        while current < self.last:
            self._values.append(current)
            current = current * self.factor
    
    @property
    def values(self):
        return self._values
    
    def init_from(self, data):

        self.load_property('factor', data)
        self.load_property('name', data)
        self.load_property('min', data)
        self.load_property('max', data)

        return self


class Experiment(Schema):

    def __init__(self):
        self.cmd = []
        self.name = None
        self.workdir = None
        self.repeat = 1
        self.max_tries = 3
        self.vars = []
    
    def init_from(self, data):

        self.load_property('name', data)
        self.load_property('workdir', data)
        self.load_property('repeat', data)
        self.load_property('max_tries', data)
        self.load_property('cmd', data, mandatory=True)

        if not isinstance(self.cmd, list):
            self.cmd = [self.cmd]
        
        if 'vars' in data:
            self.vars = []

            for data2 in data['vars']:

                if 'type' in data2:
                    if data2['type'] == 'list':
                        self.vars.append(ListVariable(data2))

                    elif data2['type'] == 'arithmetic':
                        self.vars.append(ArithmeticVariable(data2))

                    elif data2['type'] == 'geometric':
                        self.vars.append(GeometricVariable(data2))

                    else:
                        error(f"Invalid property value in Variable Schema: type={data2['type']}")
                else:
                    error('Missing mandatory property in Variable Schema: type')

        return self


def load_cluster(filepath):
    with open(filepath, "r") as fin:
        data = yaml.load(fin, Loader=yaml.FullLoader)
        return Cluster().init_from(data)


def load_experiment(filepath):
    with open(filepath, "r") as fin:
        data = yaml.load(fin, Loader=yaml.FullLoader)
        return Experiment().init_from(data)


def save_cluster(filepath:str, data:Cluster):
    with open(filepath, "w") as fout:
        yaml.dump(format_as_dict(data), fout, default_flow_style=False)


def save_experiment(filepath:str, data:Cluster):
    with open(filepath, "w") as fout:
        yaml.dump(format_as_dict(data), fout, default_flow_style=False)


def format_as_yaml(data:Schema):
    return yaml.dump(format_as_dict(data), default_flow_style=False)




from .utils import expand_path, error, warn, info, debug, critical, abort, readlines, clean_folder, estimate, human_time, quote, colors
from .schemas import Experiment, Cluster, Node, format_as_dict

from multiprocessing import Process, Queue
from subprocess import Popen, PIPE
from datetime import datetime

import hashlib
import base64
import select
import shlex
import copy
import time
import yaml
import pty
import sys
import os


KEY_SSH_ON  = b'74ffc7c4-a6ad-4315-94cb-59d045a230c0'
KEY_SSH_OFF = b'93dfc971-fa64-4beb-a24e-d8874738b9ca'
KEY_CMD_ON  = b'15e6896c-3ea7-42a0-aa32-23e2ab3c0e12'
KEY_CMD_OFF = b'e04a4348-8092-46a6-8e0c-d30d10c86fb3'

echo = lambda x: b" echo -e \"%s\"" % x.replace(b"-", b"-\b-")

ECHO_SSH_ON  = echo(KEY_SSH_ON)
ECHO_SSH_OFF = echo(KEY_SSH_OFF)
ECHO_CMD_ON  = echo(KEY_CMD_ON)
ECHO_CMD_OFF = b" echo -en \"\n $? %s\"" % KEY_CMD_OFF.replace(b"-", b"-\b-")


def combine_variables(variables, combination={}):

    if len(variables) == len(combination):
        yield copy.copy(combination)
    
    else:
        var = variables[len(combination)]
        name = var.name
        
        for value in var.values:
            combination[name] = value
            
            for tmp in combine_variables(variables, combination):
                yield tmp
            
        del combination[name]


class QueueMsg(dict):

    def __init__(self, action, source=-1):
        self.action = action
        self.source = source
        self.data = {}
    
    def __setstate__(self, state):
        self.__dict__ = state

    def __getstate__(self):
        return self.__dict__

    def __getattr__(self, key):
        return self[key]
    
    def __setattr__(self, key, value):
        self[key] = value


class Task:

    def __init__(self, 
            experiment_name, task_output_dir, work_dir, experiment_idd, combination_idd, 
            repeat_idd, task_idd, combination, cmdlines, max_tries):

        self.experiment_name = experiment_name
        self.combination_idd = combination_idd
        self.experiment_idd = experiment_idd
        self.output_dir = task_output_dir
        self.combination = combination
        self.repeat_idd = repeat_idd
        self.max_tries = max_tries
        self.work_dir = work_dir
        self.task_idd = task_idd
        self.commands = cmdlines
        self.assigned_to = None
        self.started_at = None
        self.ended_at = None
        self.duration = None
        self.success = None
        self.output = None
        self.status = None
        self.tries = 0
    
    
    def __repr__(self):
        comb = ";".join(f"{k}={v}" for k, v in self.combination.items())
        return "%d %d %d %d %s %s" % (self.experiment_idd, self.combination_idd, 
                    self.repeat_idd, self.task_idd, comb, self.commands)


class ExecutorBuilder:

    def __init__(self, executor_type, node):
        self.executor_type = executor_type
        self.node = node

    def __call__(self):
        return self.executor_type(self.node)


class BashExecutor:
    
    def __init__(self, node):

        self.is_alive = True
        self.node = node
    
    def execute(self, initrc, cmds):

        if type(cmds) is not list:
            cmds = [cmds]

        cmds = [x + b' 2>&1 ' for x in cmds]

        p1 = b" ; ".join(initrc)
        p3 = b" ; ".join(cmds)

        cmd_str = b" %s ; %s " % (p1, p3)
        cmd_str = "bash -c " + quote(cmd_str.decode('utf-8'))
        
        ps = Popen(shlex.split(cmd_str), stdout=PIPE)

        status = ps.wait()
        stdout = ps.stdout

        return (status == 0), stdout, status


class SSHExecutor:

    def __init__(self, node):

        self.node = node
        self.is_alive = False

        # Opens a pseudo-terminal
        self.master, self.slave = pty.openpty()
        self._start_bash()
        
        self.conn_string = self._build_connnection_string(self.node)

        self._connect()
    
    def _build_connnection_string(self, node):

        tokens = [b' ssh']

        if node.private_key:
            tokens.append(b' -i ')
            tokens.append(node.private_key.encode())
        
        if node.port:
            tokens.append(b' -p ')
            tokens.append(str(node.port).encode())

        tokens.append(b' -t ')
        tokens.append(node.credential.encode())

        tokens.append(b" '")
        tokens.append(ECHO_SSH_ON)
        tokens.append(b" ; bash' ; ")
        tokens.append(ECHO_SSH_OFF)
        tokens.append(b'\n')
        
        return b"".join(tokens)


    def _start_bash(self):

        debug("Starting bash")

        self.popen = Popen(
                shlex.shlex("bash"),
                preexec_fn=os.setsid,
                stdin=self.slave,
                stdout=self.slave,
                stderr=self.slave,
                universal_newlines=True)


    def _connect(self):

        conn_try = 1

        while True:
            debug("Connection attempt:", conn_try)

            os.write(self.master, self.conn_string)
            found_ssh_off = False
            found_ssh_on = False
            lines = []

            while True:
                if self.popen.poll() is not None:
                    warn("Bash has died, starting it again")
                    self._start_bash()
                
                r, _, _ = select.select([self.master], [], [])

                if not self.master in r:
                    warn("Unexpected file descriptor while stablishing connection")
                    continue
                
                for i in range(*readlines(self.master, lines, verbose=False)):
                    line = lines[i]

                    if KEY_SSH_ON in line:
                        found_ssh_on = True
                    
                    elif KEY_SSH_OFF in line:
                        found_ssh_off = True
                
                if found_ssh_on:
                    debug("SSH connection established")
                    self.is_alive = True
                    return
                
                if found_ssh_off:
                    warn("SSH connection against %s has failed, trying again" % self.node.name)
                    break

            warn("Sleeping before next try...")
            time.sleep(1)
            conn_try += 1


    def execute(self, initrc, cmds):

        if type(cmds) is not list:
            cmds = [cmds]

        lines = []
        output_start = 0
        output_end = None

        p1 = b" ; ".join(initrc)
        p2 = ECHO_CMD_ON
        p3 = b" ; ".join(cmds)
        p4 = ECHO_CMD_OFF

        cmd_str = b" %s ; %s ; %s ; %s\n" % (p1, p2, p3, p4)

        os.write(self.master, cmd_str)
        
        while True:
            if self.popen.poll() is not None:
                return False, None, None
            
            r, _, _ = select.select([self.master], [], [])

            if not self.master in r:
                warn("Unexpected file descriptor while searching for command output")
                continue

            for i in range(*readlines(self.master, lines, verbose=False)):
                line = lines[i]

                if KEY_SSH_OFF in line:
                    warn("Found KEY_SSH_OFF")
                    self.popen.kill()
                    self.is_alive = False
                    return False, None, None
                
                elif KEY_CMD_ON in line:
                    #debug("Found KEY_CMD_ON")
                    output_start = i + 1
                
                elif KEY_CMD_OFF in line and output_end is None:
                    #debug("Found KEY_CMD_OFF")
                    output_end = i
            
            if output_end is not None:
                status = lines[output_end].strip().split()[0].decode("utf-8") if output_end < len(lines) else "255"
                break

        return (status == "0"), lines[output_start:output_end], status


class WorkerProcess:

    def __init__(self, worker_idd, executor_builder, env_variables):

        self.executor_builder = executor_builder
        self.env_variables = env_variables
        self.worker_idd = worker_idd
        self.process = None
        self.queue = None


    def start(self, queue_master):

        self.queue = Queue()
        self.process = Process(target=self.run, args=(self.queue, queue_master))
        self.process.start()


    # def debug(self, *args):
    #     debug(self.worker_idd, "|", *args)


    def run(self, queue_in, queue_master):

        try:
            executor = self.executor_builder()

            msg_out = QueueMsg("ready", self.worker_idd)
            queue_master.put(msg_out)
            
            while True:
                msg_in = queue_in.get()

                if msg_in.action == "execute":
                    success = self.execute(msg_in, executor)

                    msg_out = QueueMsg("finished", self.worker_idd)
                    msg_out.success = success
                    queue_master.put(msg_out)

                    if not executor.is_alive:
                        executor = self.executor_builder()

                    msg_out = QueueMsg("ready", self.worker_idd)
                    queue_master.put(msg_out)
                
                elif msg_in.action == "terminate":
                    break
                
                else:
                    warn("Unknown action:", msg_in.action)
            
            msg_out = QueueMsg("ended", self.worker_idd)
            queue_master.put(msg_out)
        except KeyboardInterrupt:
            pass


    def execute(self, msg_in, executor):

        task:Task = msg_in.task

        # Prepare the initrc

        env_variables = copy.copy(self.env_variables)
        env_variables["PATAS_WORK_DIR"] = task.work_dir

        for k,v in task.combination.items():
            env_variables["PATAS_VAR_" + k] = str(v)

        initrc = [b"export %s=\"%s\"" % (a.encode(), b.encode()) for a, b in env_variables.items()]
        initrc.insert(0, b"cd \"%s\"" % task.work_dir.encode())

        # Prepare the command line we will execute

        cmdline = " ; ".join(task.commands).encode()

        # Execute this task

        task.started_at = datetime.now()
        task.success, task.output, task.status = executor.execute(initrc, cmdline)
        task.ended_at = datetime.now()
        task.duration = (task.ended_at - task.started_at).total_seconds()

        # Create the task folder and clean any old .done file

        done_filepath = os.path.join(task.output_dir, ".done")
        os.makedirs(task.output_dir, exist_ok=True)
        if os.path.exists(done_filepath):
            os.remove(done_filepath)

        # Dump task info
        
        info = {
            "task_id": task.task_idd,
            "repeat_id": task.repeat_idd,
            "experiment_id": task.experiment_idd,
            "experiment_name": task.experiment_name,
            "combination_id": task.combination_idd,
            "combination": task.combination,
            "started_at": task.started_at,
            "ended_at": task.ended_at,
            "duration": task.duration,
            "env_variables": env_variables,
            "max_tries": task.max_tries,
            "tries": task.tries + 1,
            "output_dir": task.output_dir,
            "work_dir": task.work_dir,
            "commands": task.commands,
            "assigned_to": task.assigned_to,
        }

        filepath = os.path.join(task.output_dir, "info.yml")
        with open(filepath, "w") as fout:
            yaml.dump(info, fout, default_flow_style=False)
        
        # Dump task output

        filepath = os.path.join(task.output_dir, "stdout")
        with open(filepath, "wb") as fout:
            fout.writelines(task.output)

        # Create .done file if the task succeded

        if task.success:
            with open(done_filepath, 'a'):
                os.utime(done_filepath, None)

        # Write output to stdout if the task failed

        else:
            warn("--- TASK %d FAILED WITH EXIT CODE %s ---" % (task.task_idd, task.status))
            for line in task.output:
                os.write(sys.stdout.fileno(), line)
            warn("--- END OF FAILED OUTPUT ---")

        return task.success


class GridExec():

    def __init__(self, task_filters, node_filters, 
            output_dir, redo_tasks, recreate, confirmed,
            experiments, clusters):

        self.output_folder = expand_path(output_dir)
        self.task_filters = task_filters
        self.node_filters = node_filters
        self.experiments = experiments
        self.redo_tasks = redo_tasks
        self.confirmed = confirmed
        self.recreate = recreate
        self.clusters = clusters
        self.workers:list[WorkerProcess] = None
        self.tasks:list[Task] = None


    def start(self):

        os.makedirs(self.output_folder, exist_ok=True)

        self._show_summary(self.experiments, self.clusters, self.confirmed)
        self._write_signature()
        self.tasks   = self._create_tasks(self.experiments, self.output_folder, self.redo_tasks, self.task_filters)
        self.workers = self._create_workers(self.clusters, self.node_filters)
        self._exec()


    def _show_summary(self, experiments, clusters, confirmed):

        # Display experiments

        print(colors.white("\n --- Experiments --- \n"))

        total_tasks = 0
        experiment: Experiment

        for experiment in experiments:
            variables = experiment.vars
            current = 1

            for var in variables:
                current *= len(var.values)
            
            current *= experiment.repeat
            total_tasks += current

            print(f"'{experiment.name}' has {len(variables)} variable(s) and produces {current} task(s):")
            for var in variables:
                print(f"    {var.name} = {str(var.values)}, len = {len(var.values)}")
        
        # Display clusters

        print(colors.white("\n --- Clusters --- \n"))

        total_nodes = 0
        total_workers = 0
        node:Node = None
        cluster:Cluster = None

        for cluster in clusters:
            total_nodes += len(cluster.nodes)
            print(f"'{cluster.name}' has {len(cluster.nodes)} node(s):")

            for node in cluster.nodes:
                total_workers += node.workers
                print(f"    '{node.name}' has {node.workers} worker(s)")
        
        # Display total numbers and estimations

        print(colors.white("\n --- Overview --- \n"))

        print(f"Redo:          {self.redo_tasks}")
        print(f"Task filters:  {self.task_filters}")
        print(f"Node filters:  {self.node_filters}")
        print(f"Output folder: {self.output_folder}")
        print(f"Recreate:      {self.recreate}")

        print()

        print(f"Total number of clusters: {len(clusters)}")
        print(f"Total number of nodes:    {total_nodes}")
        print(f"Total number of workers:  {total_workers}")
        print(f"Total number of tasks:    {total_tasks}")

        print()

        print('Estimated time to complete if each task takes:')
        print(f"    One second: { estimate(total_tasks, total_workers,        1) }")
        print(f"    One minute: { estimate(total_tasks, total_workers,       60) }")
        print(f"    One hour:   { estimate(total_tasks, total_workers,    60*60) }")
        print(f"    One day:    { estimate(total_tasks, total_workers, 60*60*24) }")

        print()

        # Confirm

        if not confirmed:
            try:
                while True:
                    option = input('Do you want to continue? [Y/n] ').strip()
                    if option in ['Y', 'y', '']:
                        break
                    elif option in ['N', 'n']:
                        sys.exit(0)
                    else:
                        print('Invalid option')
            except KeyboardInterrupt:
                sys.exit(0)

    def _write_signature(self):

        # Calculate the signature for the received configuration

        key = hashlib.md5()

        key.update(str(format_as_dict(self.experiments)).encode())

        signature = base64.b64encode(key.digest()).decode("utf-8")

        # Data to be saved into the signature file

        info = {
            "output_folder": self.output_folder,
            "task_filters": self.task_filters,
            "node_filters": self.node_filters,
            "experiments": format_as_dict(self.experiments),
            "redo_tasks": self.redo_tasks,
            "recreate": self.recreate,
            "clusters": format_as_dict(self.clusters),
            "signature": signature
        }

        # Load the previous signature file, if it exists

        info_filepath = os.path.join(self.output_folder, "info.yml")

        try:
            with open(info_filepath, "r") as fin:
                other = yaml.load(fin, Loader=yaml.FullLoader)
        except:
            other = None

        # If the previous file was loaded and it has a signature, compare it to the current signature

        if other and "signature" in other and other["signature"] != signature:
            if self.recreate:
                print("Recreating output folder...")
                clean_folder(self.output_folder)
            else:
                abort("The output folder contains a different output configuration. Provide an empty folder or add the parameter --recreate to overwrite this one")

        # Write the new signature file
            
        with open(info_filepath, "w") as fout:
            yaml.dump(info, fout, default_flow_style=False)


    def _create_tasks(self, experiments, output_folder, redo_tasks, task_filters):

        print("Creating tasks...")

        e:Experiment = None
        combination_idd = -1
        experiment_idd = -1
        task_idd = -1
        tasks = []

        for e in experiments:

            experiment_idd += 1

            experiment_filepath = os.path.join(output_folder, e.name)
            info_filepath = os.path.join(experiment_filepath, "info.yml")
            os.makedirs(experiment_filepath, exist_ok=True)

            info = {
                "experiment_name": e.name,
                "workdir": e.workdir,
                "commands": e.cmd,
                "repeat": e.repeat,
                "max_tries": e.max_tries,
                "variables": {v.name:v.values for v in e.vars}
            }

            with open(info_filepath, "w") as fout:
                yaml.dump(info, fout, default_flow_style=False)

            for combination in combine_variables(e.vars):
                combination_idd += 1

                for repeat_idd in range(e.repeat):
                    task_idd += 1

                    if task_filters and not any(a <= task_idd < b for a, b in task_filters):
                        continue

                    task_output_folder = os.path.join(experiment_filepath, str(task_idd))

                    if not redo_tasks and os.path.exists(os.path.join(task_output_folder, ".done")):
                        continue

                    cmds = [x.format(**combination) for x in e.cmd]

                    task = Task(e.name, task_output_folder, e.workdir, experiment_idd, combination_idd, repeat_idd, task_idd, combination, cmds, e.max_tries)
                    tasks.append(task)

        if not tasks:
            abort("No tasks to do.")
        
        return tasks


    def _create_workers(self, clusters, node_filters):

        print("Creating workers...")

        cluster_idd = -1
        worker_idd = -1
        node_idd = -1

        workers = []

        cluster:Cluster = None
        node:Node = None

        for cluster in clusters:
            cluster_idd += 1

            for node in cluster.nodes:
                node_idd += 1

                for _ in range(node.workers):
                    worker_idd += 1

                    if node_filters and not any(all(tag in node.tags for tag in filter) for filter in node_filters):
                        continue

                    if node.hostname in ['localhost', '127.0.0.1']:
                        builder = ExecutorBuilder(BashExecutor, node)    
                    else:
                        builder = ExecutorBuilder(SSHExecutor, node)

                    env_variables = {
                        "PATAS_CLUSTER_NAME": cluster.name,
                        "PATAS_NODE_NAME": node.name,

                        "PATAS_CLUSTER_ID": str(cluster_idd),
                        "PATAS_NODE_ID": str(node_idd),
                        "PATAS_WORKER_ID": str(worker_idd)
                    }

                    worker = WorkerProcess(worker_idd, builder, env_variables)
                    workers.append(worker)
        
        if not workers:
            abort("No workers to work.")
        
        return workers


    def _exec(self):

        self.queue    = Queue()

        self.todo     = copy.copy(self.tasks)
        self.doing    = []
        self.done     = []
        self.given_up = []

        self.idle     = []
        self.ended    = []

        len_todo      = len(str(len(self.todo)))
        len_workers   = len(str(len(self.workers)))
        
        # Start workers

        print()
        info("Starting %d worker(s)" % len(self.workers))

        for worker in self.workers:
            worker.start(self.queue)
        
        info("Worker(s) started")

        # Main loop

        main_loop_started_at = datetime.now()

        try:

            print()
            info("Starting main loop")

            while self.todo or self.doing:

                msg_in = self.queue.get()

                l1 = str(len(self.todo    ))
                l2 = str(len(self.doing   ))
                l3 = str(len(self.done    ))
                l4 = str(len(self.given_up))

                d  = colors.gray("|%s|" % str(datetime.now()))
                l1 = colors.white(" " * (len_todo    - len(l1)) + l1 + " |")
                l2 = colors.white(" " * (len_workers - len(l2)) + l2 + " |")
                l3 = colors.green(" " * (len_todo    - len(l3)) + l3 + " |")
                l4 = colors.red  (" " * (len_todo    - len(l4)) + l4 + " |")

                print(f"{d} {l1} {l2} {l3} {l4}")

                if msg_in.action == "ready":
                    self._ready(msg_in)
                
                elif msg_in.action == "finished":
                    self._finished(msg_in)
                
                else:
                    warn("Unknown action:", msg_in.action)
            
            info("Main loop completed")

        except KeyboardInterrupt:

            print("Operation interrupted")
            return

        main_loop_ended_at = datetime.now()
        main_loop_duration = (main_loop_ended_at - main_loop_started_at).total_seconds()
        
        # Terminate workers

        terminate_loop_started_at = datetime.now()

        try:
            print()
            info("Releasing workers...")

            # Sending TERMINATE signal

            for worker in self.workers:
                msg = QueueMsg("terminate")
                worker.queue.put(msg)
            
            # Waiting for ENDED signal

            while len(self.workers) != len(self.ended):

                msg = self.queue.get()

                if msg.action == "ended":

                    self.ended.append(msg.source)

                    d = colors.gray("|%s|" % str(datetime.now()))

                    l1 = str(len(self.ended))
                    l2 = str(len(self.workers))

                    l1 = " " * (len_workers - len(l1)) + l1
                    l2 = " " * (len_workers - len(l2)) + l2

                    print("%s %sEnded %s / %s%s" % (d, colors.WHITE, l1, l2, colors.RESET))
                
                else:
                    #debug("Ignoring action %s from %s, execution is ending" % (msg.source, msg.action))
                    pass

            info("All workers are resting.")
            
        except KeyboardInterrupt:

            print("Operation interrupted")
            return

        terminate_loop_ended_at = datetime.now()
        terminate_loop_duration = (terminate_loop_ended_at - terminate_loop_started_at).total_seconds()
        
        # Display execution summary

        print(colors.white("\n --- Execution Summary --- \n"))
        print(f"    Time to execute tasks: {human_time(main_loop_duration)}")
        print(f"    Time to terminate workers: {human_time(terminate_loop_duration)}")
        print(f"    Tasks requested: {len(self.tasks)}")
        print(f"    Tasks completed: {len(self.done)}")
        print(f"    Tasks given up: {len(self.given_up)}")
        print()

    def _ready(self, msg_in):

        if self.todo:
            msg_out = QueueMsg("execute")
            msg_out.task = self.todo.pop()
            msg_out.task.assigned_to = msg_in.source

            self.doing.append(msg_out.task)
            self.workers[msg_in.source].queue.put(msg_out)
        
        else:
            self.idle.append(msg_in.source)


    def _finished(self, msg_in):

        for i, x in enumerate(self.doing):
            if x.assigned_to == msg_in.source:
                break
        
        if i == len(self.doing):
            warn("Received finished msg but the task was not found inside the doing list")
            return
        
        task:Task = self.doing[i]
        del self.doing[i]
        task.tries += 1

        if msg_in.success:
            self.done.append(task)
            return

        if task.tries >= task.max_tries:
            critical(f"Giving up on task {task.task_idd}, max_tries reached.")
            self.given_up.append(task)
            return
        
        if not self.idle:
            self.todo.append(task)
            return

        target = self.idle.pop()

        msg_out = QueueMsg("execute")
        msg_out.task = task
        msg_out.task.assigned_to = target

        self.workers[target].queue.put(msg_out)



name = "patas"
version = "2.0.1"
author = "Diego Souza"
author_email = "diegofpsouza+patas@gmail.com"
url = "https://github.com/diegofps/patas"
description = "Parallelize any program and aggregate results from stdout."




import sys

LOG_LEVEL = 2

def debug(*args):
    if LOG_LEVEL <= 0:
        print("DEBUG:", *args)

def info(*args):
    if LOG_LEVEL <= 1:
        print("INFO:", *args)

def warn(*args):
    if LOG_LEVEL <= 2:
        print("WARN:", *args)

def error(*args):
    if LOG_LEVEL <= 3:
        print("ERROR:", *args)
    sys.exit(1)



from subprocess import Popen, PIPE

import shutil
import shlex
import glob
import sys
import re
import os



class colors:
    RESET="\033[0m"

    RESET     = "\033[0m"

    BRIGHTER  = "\033[1m"
    DARKER    = "\033[2m"
    ITALIC    = "\033[3m"
    UNDERLINE = "\033[4m"
    BLINKING  = "\033[5m"
    REVERSE   = "\033[7m"
    INVISIBLE = "\033[8m"
    CROSSING  = "\033[9m"

    GRAY    = "\033[90m"
    RED     = "\033[91m"
    GREEN   = "\033[92m"
    YELLOW  = "\033[93m"
    BLUE    = "\033[94m"
    PURPLE  = "\033[95m"
    CYAN    = "\033[96m"
    WHITE   = "\033[97m"

    BOLD_GRAY   = "\033[1;30m"
    BOLD_RED    = "\033[1;31m"
    BOLD_GREEN  = "\033[1;32m"
    BOLD_YELLOW = "\033[1;33m"
    BOLD_BLUE   = "\033[1;34m"
    BOLD_PURPLE = "\033[1;35m"
    BOLD_CYAN   = "\033[1;36m"
    BOLD_WHITE  = "\033[1;37m"

    BG_GRAY   = "\033[40m"
    BG_RED    = "\033[41m"
    BG_GREEN  = "\033[42m"
    BG_YELLOW = "\033[43m"
    BG_BLUE   = "\033[44m"
    BG_PURPLE = "\033[45m"
    BG_CYAN   = "\033[46m"
    BG_WHITE  = "\033[47m"

    @staticmethod
    def gray(str):
        return colors.GRAY + str + colors.RESET
    
    @staticmethod
    def red(str):
        return colors.RED + str + colors.RESET
    
    @staticmethod
    def green(str):
        return colors.GREEN + str + colors.RESET

    @staticmethod
    def yellow(str):
        return colors.YELLOW + str + colors.RESET
        
    @staticmethod
    def blue(str):
        return colors.BLUE + str + colors.RESET
    
    @staticmethod
    def purple(str):
        return colors.PURPLE + str + colors.RESET
    
    @staticmethod
    def cyan(str):
        return colors.CYAN + str + colors.RESET
    
    @staticmethod
    def white(str):
        return colors.WHITE + str + colors.RESET

    @staticmethod
    def normal(str):
        return colors.RESET + str + colors.RESET


def readlines(master, lines, verbose=False):
    
    o = os.read(master, 10240)

    if verbose:
        os.write(sys.stdout.fileno(), o)

    if not o:
        return 0, 0
    
    last = 0
    i = 0

    linebreak = ord('\n')
    result = []

    while i != len(o):
        if o[i] == linebreak:
            result.append(o[last:i+1])
            last = i + 1
        
        i += 1
    
    result.append(o[last:i])

    start_search = max(len(lines) - 1, 0)

    if lines:
        lines[-1] = lines[-1] + result[0]
        lines.extend(result[1:])
    else:
        lines.extend(result)
    
    return start_search, len(lines)


def quote(str):
    return '"' + re.sub(r'([\'\"\\])', r'\\\1', str) + '"'


def quote_single(str):
    return "'" + re.sub(r'([\'\\])', r'\\\1', str) + "'"


def run(cmds, write=None, read=False, suppressInterruption=False, suppressError=False, verbose=False):
    if type(cmds) is not list:
        cmds = [cmds]
    
    if not cmds:
        return
    
    if verbose:
        print("Executing command:\n" + " | ".join(cmds))

    try:
        if write and read:
            args = shlex.split(cmds[0])
            ps = [Popen(args, stdout=PIPE, stdin=PIPE)]

            for i in range(1, len(cmds)):
                args = shlex.split(cmds[i])
                previous = ps[i-1]
                current = Popen(args, stdout=PIPE, stdin=previous.stdout)
                ps.append(current)

            front = ps[0]
            for line in write:
                front.stdin.write(line.encode())
            front.stdin.close()

            back = ps[-1]
            lines = [line.decode("utf-8") for line in back.stdout]
            status = back.wait()

            if not suppressError and status != 0:
                error("Failed to execute command:\n" + " | ".join(cmds))
            
            return status,lines

        elif write:
            args = shlex.split(cmds[0])

            if len(cmds) == 1:
                ps = [Popen(args, stdin=PIPE)]
            else:
                ps = [Popen(args, stdout=PIPE, stdin=PIPE)]

            final = len(cmds) - 1
            for i in range(1, final):
                args = shlex.split(cmds[i])
                previous = ps[i-1]

                if i + 1 == final:
                    ps.append(Popen(args, stdin=previous.stdout))
                else:
                    ps.append(Popen(args, stdout=PIPE, stdin=previous.stdout))

            front = ps[0]
            for line in write:
                front.stdin.write(line.encode())
            front.stdin.close()

            status = ps[-1].wait()

            if not suppressError and status != 0:
                error("Failed to execute command:\n" + " | ".join(cmds))
            
            return status, None

        elif read:
            args = shlex.split(cmds[0])
            ps = [Popen(args, stdout=PIPE)]

            for i in range(1, len(cmds)):
                args = shlex.split(cmds[i])
                previous = ps[i-1]
                current = Popen(args, stdout=PIPE, stdin=previous.stdout)
                ps.append(current)

            back = ps[-1]
            lines = [line.decode("utf-8") for line in back.stdout]
            status = back.wait()

            if not suppressError and status != 0:
                error("Failed to execute command:\n" + " | ".join(cmds))
            
            return status, lines

        elif len(cmds) == 1:
            status = os.system(cmds[0])

            if not suppressError and status != 0:
                error("Failed to execute command:\n" + " | ".join(cmds))
            
            return status, None
        
        else:
            args = shlex.split(cmds[0])
            ps = [Popen(args, stdout=PIPE)]

            for i in range(1, len(cmds)-1):
                args = shlex.split(cmds[i])
                previous = ps[i-1]
                current = Popen(args, stdout=PIPE, stdin=previous.stdout)
                ps.append(current)

            args = shlex.split(cmds[-1])
            previous = ps[-1]
            current = Popen(args, stdin=previous.stdout)
            ps.append(current)

            status = ps[-1].wait()
            
            if not suppressError and status != 0:
                error("Failed to execute command:\n" + " | ".join(cmds))
            
            return status, None

    except KeyboardInterrupt as e:
        if not suppressInterruption:
            raise e


def abort(*args):
    if args:
        msg = colors.PURPLE + ' '.join([str(x) for x in args]) + colors.RESET
        print(msg, file=sys.stderr)
    exit(1)


def error(*args):
    if args:
        raise PatasError(" ".join(args))
    else:
        raise PatasError()


def critical(*args):
    print(colors.RED + "|CRITICAL| " + " ".join([str(v) for v in args]) + colors.RESET)


def warn(*args):
    print(colors.YELLOW + "|WARNING| " + " ".join([str(v) for v in args]) + colors.RESET)


def info(*args):
    print(colors.GREEN + "|INFO| " + " ".join([str(v) for v in args]) + colors.RESET)


def debug(*args):
    print(colors.CYAN + "|DEBUG| " + " ".join([str(v) for v in args]) + colors.RESET)


def filename(filepath):
    return os.path.splitext(os.path.basename(filepath))[0]


def expand_path(filepath):
    if not filepath:
        return filepath
    
    if filepath.startswith("/"):
        return filepath
    
    if filepath.startswith("./") or filepath.startswith("../"):
        return os.path.abspath(filepath)
    
    if filepath.startswith("~"):
        return os.path.expanduser(filepath)
    
    return os.path.abspath(os.path.join(".", filepath))


class PatasError(Exception):
    def __init__(self, message=None):
        self.message = message


def clean_folder(folderpath):
    
    if os.path.exists(folderpath):
        if os.path.isdir(folderpath):
            for filepath in glob.glob(os.path.join(folderpath, '*')):
                debug("  Removing:", filepath)
                if os.path.isdir(filepath):
                    shutil.rmtree(filepath)
                else:
                    os.remove(filepath)
        else:
            os.remove(folderpath)
    else:
        os.makedirs(folderpath, exist_ok=True)


def human_time(seconds):

    v = seconds

    if v < 60:
        return f"{v} seconds"
    v /= 60

    if v < 60:
        return f"{v} minutes"
    v /= 60

    if v < 24:
        return f"{v} hours"
    v /= 24

    if v < 7:
        return f"{v} days"
    v /= 7

    if v < 52:
        return f"{v} weeks ~ {v / 4.0} months"
    v /= 52

    if v < 100:
        return f"{v} years"
    v /= 100

    if v < 10:
        return f"{v} centuries"
    v /= 10

    return f"{v} milleniums"


def estimate(tasks, workers, seconds):
    
    return human_time(tasks / workers * seconds)


def node_cpu_count(user, hostname, port, default):

    if hostname in ['localhost', '127.0.0.1']:
        status, output = run('nproc', read=True)

    else:
        cmd = [' ssh ']

        if user:
            cmd.append(user)
            cmd.append('@')

        cmd.append(hostname)

        if port:
            cmd.append(' -p')
            cmd.append(str(port))
        
        cmd.append(" 'nproc'")

        status, output = run("".join(cmd), read=True)

    if status != 0:
        return default

    try:
        return int(output[0])
    except ValueError:
        return default
                


from patas.utils import quote_single, run, abort

from glob import glob

import os


class QueryEngine:

    def __init__(self, patas_folder):
        
        self.tables = []

        for folderpath in glob(os.path.join(patas_folder, '*')):
            if os.path.isdir(folderpath):
                experiment_name = os.path.basename(folderpath)
                table_filepath = os.path.join(folderpath, experiment_name + '.csv')
                if os.path.exists(table_filepath):
                    self.tables.append(f'"{table_filepath}"')

    def query(self, query, pretty_print):
        

        tables = " ".join(self.tables)

        if pretty_print:
            cmd = f"csvsql -d ';' --query {quote_single(query)} {tables} 2> /dev/null | csvlook"
        else:
            cmd = f"csvsql -d ';' --query {quote_single(query)} {tables} 2> /dev/null"
        
        status, _ = run(cmd)

        if status != 0:
            abort("Could not call csvkit, is it accessible in $PATH?")
        

